/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __BasicCommand_H__
#define __BasicCommand_H__


#include <list>
#include <cstring>
#include <string>
#include "Types.h"
#include "Utilities.h"
#include "Geometry.h"

class Agent;

//==============================================================================
struct CommandInfo
{
	CommandType mType;
	int mLevel;
	int mMinVer;
	int mMaxVer;
	bool mMutex;
	Time mTime;
	double mDist;
	double mPower;
	double mAngle;
	Vector mMovePos;
	ViewWidth mViewWidth;
    std::string mString;

	CommandInfo()
	{
		mType = CT_None;
	}

	const double & GetPower() const { return mPower; }
	const double & GetAngle() const { return mAngle; }
	const Vector & GetMovePos() const { return mMovePos; }
};


//==============================================================================
class BasicCommand
{
public:
	BasicCommand(const Agent & agent): mAgent(agent) {}
	virtual ~BasicCommand() {}

	bool Execute(std::list<CommandInfo> &command_queue);

	const double & GetPower() const { return mCommandInfo.GetPower(); }
	const double & GetAngle() const { return mCommandInfo.GetAngle(); }
	const Vector & GetMovePos() const { return mCommandInfo.GetMovePos(); }

protected:
	const Agent     & mAgent;
	CommandInfo mCommandInfo; // Evaluated in Plan()pushed back in Execute()
};


/**
* serverԽյԭ࣬˳SoccerServer - Player::parseCommand()е˳һ
* The followings are the basic actions can be operated by server immediately. 
* Ordered by SoccerServer - Player::parseCommand().
*/

//==============================================================================
class Turn : public BasicCommand
{
public:
	Turn(const Agent & agent);
	~Turn() {}

	void Plan(double moment);
};


//==============================================================================
class Dash : public BasicCommand
{
public:
	Dash(const Agent & agent);
	~Dash() {}

	void Plan(double power, AngleDeg dir);
};


//==============================================================================
class TurnNeck : public BasicCommand
{
public:
	TurnNeck(const Agent & agent);
	~TurnNeck() {}

	void Plan(double moment);
};


//==============================================================================
class Say : public BasicCommand
{
public:
	Say(const Agent & agent);
	~Say() {}

	void Plan(std::string msg);
};


//==============================================================================
class Attentionto : public BasicCommand
{
public:
	Attentionto(const Agent & agent);
	~Attentionto() {}

	void Plan(bool on, Unum num = 0);
};


//==============================================================================
class Kick : public BasicCommand
{
public:
	Kick(const Agent & agent);
	~Kick() {}

	void Plan(double power, double dir);
};


//==============================================================================
class Tackle : public BasicCommand
{
public:
	Tackle(const Agent & agent);
	~Tackle() {}

	void Plan(double dir);
};


//==============================================================================
class Pointto : public BasicCommand
{
public:
	Pointto(const Agent & agent);
	~Pointto() {}

	void Plan(bool on, double dist = 0.0, double dir = 0.0);
};


//==============================================================================
class Catch : public BasicCommand
{
public:
	Catch(const Agent & agent);
	~Catch() {}

	void Plan(double dir);
};


//==============================================================================
class Move : public BasicCommand
{
public:
	Move(const Agent & agent);
	~Move() {}

	void Plan(Vector pos);
};

//==============================================================================
class ChangeView : public BasicCommand
{
public:
	ChangeView(const Agent & agent);
	~ChangeView() {}

	void Plan(ViewWidth view_width);
};


//==============================================================================
class Compression : public BasicCommand
{
public:
	Compression(const Agent & agent);
	~Compression() {}

	void Plan(int level);
};


//==============================================================================
class SenseBody : public BasicCommand
{
public:
	SenseBody(const Agent & agent);
	~SenseBody() {}

	void Plan();
};


//==============================================================================
class Score : public BasicCommand
{
public:
	Score(const Agent & agent);
	~Score() {}

	void Plan();
};


//==============================================================================
class Bye : public BasicCommand
{
public:
	Bye(const Agent & agent);
	~Bye() {}

	void Plan();
};


//==============================================================================
class Done : public BasicCommand
{
public:
	Done(const Agent & agent);
	~Done() {}

	void Plan();
};


//==============================================================================
class Clang : public BasicCommand
{
public:
	Clang(const Agent & agent);
	~Clang() {}

	void Plan(int min_ver, int max_ver);
};


//==============================================================================
class Ear : public BasicCommand
{
public:
	Ear(const Agent & agent);
	~Ear() {}

	void Plan(bool on, bool our_side, EarMode ear_mode = EM_All);
};


//==============================================================================
class SynchSee : public BasicCommand
{
public:
	SynchSee(const Agent & agent);
	~SynchSee() {}

	void Plan();
};

//==============================================================================
class ChangePlayerType : public BasicCommand
{
public:
	ChangePlayerType(const Agent & agent);
	~ChangePlayerType() {}

	void Plan(Unum num, int player_type);
};

#endif

